import halide as hl
import numpy as np
import gc
import sys


def test_ndarray_to_buffer(reverse_axes=True):
    a0 = np.ones((200, 300), dtype=np.int32)

    # Buffer always shares data (when possible) by default,
    # and maintains the shape of the data source. (note that
    # the ndarray is col-major by default!)
    b0 = hl.Buffer(a0, "float32_test_buffer", reverse_axes)
    assert b0.type() == hl.Int(32)
    assert b0.name() == "float32_test_buffer"
    assert b0.all_equal(1)

    if reverse_axes:
        assert b0.dim(0).min() == 0
        assert b0.dim(0).max() == 299
        assert b0.dim(0).extent() == 300
        assert b0.dim(0).stride() == 1

        assert b0.dim(1).min() == 0
        assert b0.dim(1).max() == 199
        assert b0.dim(1).extent() == 200
        assert b0.dim(1).stride() == 300

        a0[12, 34] = 56
        assert b0[34, 12] == 56

        b0[56, 34] = 12
        assert a0[34, 56] == 12
    else:
        assert b0.dim(0).min() == 0
        assert b0.dim(0).max() == 199
        assert b0.dim(0).extent() == 200
        assert b0.dim(0).stride() == 300

        assert b0.dim(1).min() == 0
        assert b0.dim(1).max() == 299
        assert b0.dim(1).extent() == 300
        assert b0.dim(1).stride() == 1

        a0[12, 34] = 56
        assert b0[12, 34] == 56

        b0[56, 34] = 12
        assert a0[56, 34] == 12


def test_buffer_to_ndarray(reverse_axes=True):
    buf0 = hl.Buffer(hl.Int(16), [4, 6])
    assert buf0.type() == hl.Int(16)
    buf0.fill(0)
    buf0[1, 2] = 42
    assert buf0[1, 2] == 42

    # This is subtle: the default behavior when converting
    # a Buffer to an np.array (or ndarray, etc) is to reverse the
    # order of the axes, since Halide prefers column-major and
    # the rest of Python prefers row-major. By calling reverse_axes()
    # before that conversion, we end up doing a *double* reverse, i.e,
    # not reversing at all. So the 'not' here is correct.
    buf = buf0.reverse_axes() if not reverse_axes else buf0

    # Should share storage with buf
    array_shared = np.array(buf, copy=False)
    assert array_shared.dtype == np.int16
    if reverse_axes:
        assert array_shared.shape == (6, 4)
        assert array_shared[2, 1] == 42
    else:
        assert array_shared.shape == (4, 6)
        assert array_shared[1, 2] == 42

    # Should *not* share storage with buf
    array_copied = np.array(buf, copy=True)
    assert array_copied.dtype == np.int16
    if reverse_axes:
        assert array_copied.shape == (6, 4)
        assert array_copied[2, 1] == 42
    else:
        assert array_copied.shape == (4, 6)
        assert array_copied[1, 2] == 42

    # Should affect array_shared but not array_copied
    buf0[1, 2] = 3
    if reverse_axes:
        assert array_shared[2, 1] == 3
        assert array_copied[2, 1] == 42
    else:
        assert array_shared[1, 2] == 3
        assert array_copied[1, 2] == 42

    # Ensure that Buffers that have nonzero mins get converted correctly,
    # since the Python Buffer Protocol doesn't have the 'min' concept
    cropped_buf0 = buf0.copy()
    cropped_buf0.crop(dimension=0, min=1, extent=2)
    cropped_buf = cropped_buf0.reverse_axes() if not reverse_axes else cropped_buf0

    # Should share storage with cropped (and buf)
    cropped_array_shared = np.array(cropped_buf, copy=False)
    assert cropped_array_shared.dtype == np.int16
    if reverse_axes:
        assert cropped_array_shared.shape == (6, 2)
        assert cropped_array_shared[2, 0] == 3
    else:
        assert cropped_array_shared.shape == (2, 6)
        assert cropped_array_shared[0, 2] == 3

    # Should *not* share storage with anything
    cropped_array_copied = np.array(cropped_buf, copy=True)
    assert cropped_array_copied.dtype == np.int16
    if reverse_axes:
        assert cropped_array_copied.shape == (6, 2)
        assert cropped_array_copied[2, 0] == 3
    else:
        assert cropped_array_copied.shape == (2, 6)
        assert cropped_array_copied[0, 2] == 3

    cropped_buf0[1, 2] = 5
    assert cropped_buf0[1, 2] == 5
    if reverse_axes:
        assert cropped_buf[1, 2] == 5
        assert cropped_array_shared[2, 0] == 5
        assert cropped_array_copied[2, 0] == 3
    else:
        assert cropped_buf[2, 1] == 5
        assert cropped_array_shared[0, 2] == 5
        assert cropped_array_copied[0, 2] == 3


def _assert_fn(e):
    assert e


def _is_64bits():
    return sys.maxsize > 2**32


def test_for_each_element():
    buf = hl.Buffer(hl.Float(32), [3, 4])
    for x in range(3):
        for y in range(4):
            buf[x, y] = x + y
    # Can't use 'assert' in a lambda, but can call a fn that uses it.
    buf.for_each_element(
        lambda pos, buf=buf: _assert_fn(buf[pos[0], pos[1]] == pos[0] + pos[1])
    )


def test_fill_all_equal():
    buf = hl.Buffer(hl.Int(32), [3, 4])
    buf.fill(3)
    assert buf.all_equal(3)
    buf[1, 2] = 4
    assert not buf.all_equal(3)


def test_bufferinfo_sharing():
    # Don't bother testing this on 32-bit systems (our "huge" size is too large there)
    if not _is_64bits():
        print("skipping test_bufferinfo_sharing()")
        return

    # Torture-test to ensure that huge Python Buffer Protocol allocations are properly
    # shared (rather than copied), and also that the lifetime is held appropriately.
    a0 = np.ones((20000, 30000), dtype=np.int32)
    b0 = hl.Buffer(a0)
    del a0
    for i in range(200):
        b1 = hl.Buffer(b0)
        b0 = b1
        b1 = None
        gc.collect()

    b0[56, 34] = 12
    assert b0[56, 34] == 12


def test_float16():
    array_in = np.zeros((256, 256, 3), dtype=np.float16, order="F")
    hl_img = hl.Buffer(array_in)
    array_out = np.array(hl_img, copy=False)


# TODO: https://github.com/halide/Halide/issues/6849
# def test_bfloat16():
#     try:
#         from tensorflow.python.lib.core import _pywrap_bfloat16
#         bfloat16 = _pywrap_bfloat16.TF_bfloat16_type()
#         array_in = np.zeros((256, 256, 3), dtype=bfloat16, order='F')
#         hl_img = hl.Buffer(array_in)
#         array_out = np.array(hl_img, copy = False)
#     except ModuleNotFoundError as e:
#         print("skipping test_bfloat16() because tensorflow was not found: %s" % str(e))
#         return
#     else:
#         assert False, "This should not happen"


def test_int64():
    array_in = np.zeros((256, 256, 3), dtype=np.int64, order="F")
    hl_img = hl.Buffer(array_in)
    array_out = np.array(hl_img, copy=False)


def test_make_interleaved():
    w = 7
    h = 13
    c = 3

    b = hl.Buffer.make_interleaved(type=hl.UInt(8), width=w, height=h, channels=c)

    assert b.dim(0).min() == 0
    assert b.dim(0).extent() == w
    assert b.dim(0).stride() == c

    assert b.dim(1).min() == 0
    assert b.dim(1).extent() == h
    assert b.dim(1).stride() == w * c

    assert b.dim(2).min() == 0
    assert b.dim(2).extent() == c
    assert b.dim(2).stride() == 1

    a = np.array(b, copy=False)
    # NumPy shape order is opposite that of Halide shape order
    assert a.shape == (c, h, w)
    assert a.strides == (1, w * c, c)
    assert a.dtype == np.uint8


def test_interleaved_ndarray():
    w = 7
    h = 13
    c = 3

    a = np.ndarray(dtype=np.uint8, shape=(w, h, c), strides=(c, w * c, 1))

    assert a.shape == (w, h, c)
    assert a.strides == (c, w * c, 1)
    assert a.dtype == np.uint8

    b = hl.Buffer(a)
    assert b.type() == hl.UInt(8)

    assert b.dim(0).min() == 0
    assert b.dim(0).extent() == c
    assert b.dim(0).stride() == 1

    assert b.dim(1).min() == 0
    assert b.dim(1).extent() == h
    assert b.dim(1).stride() == w * c

    assert b.dim(2).min() == 0
    assert b.dim(2).extent() == w
    assert b.dim(2).stride() == c


def test_reorder():
    W = 7
    H = 5
    C = 3
    Z = 2

    a = hl.Buffer(type=hl.UInt(8), sizes=[W, H, C], storage_order=[2, 0, 1])
    assert a.dim(0).extent() == W
    assert a.dim(1).extent() == H
    assert a.dim(2).extent() == C
    assert a.dim(2).stride() == 1
    assert a.dim(0).stride() == C
    assert a.dim(1).stride() == W * C

    b = hl.Buffer(hl.UInt(8), [W, H, C, Z], [2, 3, 0, 1])
    assert b.dim(0).extent() == W
    assert b.dim(1).extent() == H
    assert b.dim(2).extent() == C
    assert b.dim(3).extent() == Z
    assert b.dim(2).stride() == 1
    assert b.dim(3).stride() == C
    assert b.dim(0).stride() == C * Z
    assert b.dim(1).stride() == W * C * Z

    b2 = hl.Buffer(hl.UInt(8), [C, Z, W, H])
    assert b.dim(0).extent() == b2.dim(2).extent()
    assert b.dim(1).extent() == b2.dim(3).extent()
    assert b.dim(2).extent() == b2.dim(0).extent()
    assert b.dim(3).extent() == b2.dim(1).extent()
    assert b.dim(0).stride() == b2.dim(2).stride()
    assert b.dim(1).stride() == b2.dim(3).stride()
    assert b.dim(2).stride() == b2.dim(0).stride()
    assert b.dim(3).stride() == b2.dim(1).stride()

    b2.transpose([2, 3, 0, 1])
    assert b.dim(0).extent() == b2.dim(0).extent()
    assert b.dim(1).extent() == b2.dim(1).extent()
    assert b.dim(2).extent() == b2.dim(2).extent()
    assert b.dim(3).extent() == b2.dim(3).extent()
    assert b.dim(0).stride() == b2.dim(0).stride()
    assert b.dim(1).stride() == b2.dim(1).stride()
    assert b.dim(2).stride() == b2.dim(2).stride()
    assert b.dim(3).stride() == b2.dim(3).stride()


def test_overflow():
    # Don't bother testing this on 32-bit systems (our "huge" size is too large there)
    if not _is_64bits():
        print("skipping test_overflow()")
        return

    # size = INT_MAX
    w_intmax = 0x7FFFFFFF

    # When size == INT_MAX, we should not emit error
    size_intmax = np.ndarray(dtype=np.uint8, shape=(w_intmax))
    hl.Buffer(size_intmax)

    # size = INT_MAX + 1
    w_over_intmax = 0x7FFFFFFF + 1

    # We should emit the error when the size > INT_MAX
    size_over_intmax = np.ndarray(dtype=np.uint8, shape=(w_over_intmax))
    try:
        hl.Buffer(size_over_intmax)
    except ValueError as e:
        assert "Out of range dimensions in buffer conversion" in str(e)


def test_buffer_to_str():
    b = hl.Buffer()
    assert str(b) == "<undefined halide.Buffer>"
    b = hl.Buffer(hl.Int(32), [128, 256])
    assert str(b) == "<halide.Buffer of type int32 shape:[[0,128,1],[0,256,128]]>"


def test_scalar_buffers():
    buf = hl.Buffer.make_scalar(hl.Float(32))

    assert buf.dimensions() == 0

    buf.fill(0)
    buf[()] = 2.5

    assert buf[()] == 2.5

    buf.fill(32)
    assert buf[()] == 32


def test_oob():
    buf = hl.Buffer(hl.Int(16), [4, 6])
    buf.fill(0)

    # getitem below min
    try:
        print(buf[-1, 2])
    except IndexError as e:
        assert "index -1 is out of bounds for axis 0 with min=0, extent=4" in str(e)
    else:
        assert False, "Did not see expected exception!"

    # getitem above max
    try:
        print(buf[1, 6])
    except IndexError as e:
        assert "index 6 is out of bounds for axis 1 with min=0, extent=6" in str(e)

    # setitem below min
    try:
        buf[-1, 2] = 42
    except IndexError as e:
        assert "index -1 is out of bounds for axis 0 with min=0, extent=4" in str(e)
    else:
        assert False, "Did not see expected exception!"

    # setitem above max
    try:
        buf[1, 6] = 42
    except IndexError as e:
        assert "index 6 is out of bounds for axis 1 with min=0, extent=6" in str(e)


if __name__ == "__main__":
    test_make_interleaved()
    test_interleaved_ndarray()
    test_ndarray_to_buffer(reverse_axes=True)
    test_ndarray_to_buffer(reverse_axes=False)
    test_buffer_to_ndarray(reverse_axes=True)
    test_buffer_to_ndarray(reverse_axes=False)
    test_for_each_element()
    test_fill_all_equal()
    test_bufferinfo_sharing()
    # TODO: https://github.com/halide/Halide/issues/6849
    # test_bfloat16()
    test_float16()
    test_int64()
    test_reorder()
    test_overflow()
    test_buffer_to_str()
    test_scalar_buffers()
    test_oob()



#!/usr/bin/python3

# Halide tutorial lesson 11.

# This lesson demonstrates how to use Halide as a cross-compiler.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_11_cross_compilation
# in a shell with the current directory at python_bindings/

import halide as hl
from struct import unpack


def main():

    # We'll define the simple one-stage pipeline that we used in lesson 10.
    brighter = hl.Func("brighter")
    x, y = hl.Var("x"), hl.Var("y")

    # Declare the arguments.
    offset = hl.Param(hl.UInt(8))
    input = hl.ImageParam(hl.UInt(8), 2)
    args = [input, offset]

    # Define the hl.Func.
    brighter[x, y] = input[x, y] + offset

    # Schedule it.
    brighter.vectorize(x, 16).parallel(y)

    # The following line is what we did in lesson 10. It compiles an
    # object file suitable for the system that you're running this
    # program on.  For example, if you compile and run this file on
    # 64-bit linux on an x86 cpu with sse4.1, then the generated code
    # will be suitable for 64-bit linux on x86 with sse4.1.
    brighter.compile_to_file("lesson_11_host", args, "lesson_11_host")

    # We can also compile object files suitable for other cpus and
    # operating systems. You do this with an optional third argument
    # to compile_to_file which specifies the target to compile for.

    create_android = True
    create_windows = True
    create_ios = True

    if create_android:
        # Let's use this to compile a 32-bit arm android version of this code:
        target = hl.Target()
        target.os = hl.TargetOS.Android  # The operating system
        target.arch = hl.TargetArch.ARM  # The CPU architecture
        target.bits = 32              # The bit-width of the architecture
        arm_features = []             # A list of features to set
        target.set_features(arm_features)
        # Pass the target as the last argument.
        brighter.compile_to_file("lesson_11_arm_32_android", args,
                                 "lesson_11_arm_32_android", target)

    if create_windows:
        # And now a Windows object file for 64-bit x86 with AVX and SSE 4.1:
        target = hl.Target()
        target.os = hl.TargetOS.Windows
        target.arch = hl.TargetArch.X86
        target.bits = 64
        target.set_features([hl.TargetFeature.AVX, hl.TargetFeature.SSE41])
        brighter.compile_to_file("lesson_11_x86_64_windows", args,
                                 "lesson_11_x86_64_windows", target)

    if create_ios:
        # And finally an iOS mach-o object file for one of Apple's 32-bit
        # ARM processors - the A6. It's used in the iPhone 5. The A6 uses
        # a slightly modified ARM architecture called ARMv7s. We specify
        # this using the target features field.  Support for Apple's
        # 64-bit ARM processors is very new in llvm, and still somewhat
        # flaky.
        target = hl.Target()
        target.os = hl.TargetOS.IOS
        target.arch = hl.TargetArch.ARM
        target.bits = 32
        target.set_features([hl.TargetFeature.ARMv7s])
        brighter.compile_to_file("lesson_11_arm_32_ios", args,
                                 "lesson_11_arm_32_ios", target)

    # Now let's check these files are what they claim, by examining
    # their first few bytes.

    if create_android:
        # 32-arm android object files start with the magic bytes:
        # uint8_t []
        arm_32_android_magic = [0x7f, ord('E'), ord('L'), ord('F'),  # ELF format
                                1,       # 32-bit
                                1,       # 2's complement little-endian
                                1]       # Current version of elf

        length = len(arm_32_android_magic)
        f = open("lesson_11_arm_32_android.o", "rb")
        try:
            header_bytes = f.read(length)
        except:
            assert False, "Android object file not generated"
        f.close()

        header = list(unpack("B" * length, header_bytes))
        assert header == arm_32_android_magic, \
            "Unexpected header bytes in 32-bit arm object file: " + \
            str([x == y for x, y in zip(header, arm_32_android_magic)])

    if create_windows:
        # 64-bit windows object files start with the magic 16-bit value 0x8664
        # (presumably referring to x86-64)
        # uint8_t  []
        win_64_magic = [0x64, 0x86]

        f = open("lesson_11_x86_64_windows.obj", "rb")
        try:
            header_bytes = f.read(2)
        except:
            assert False, "Windows object file not generated"
        f.close()

        header = list(unpack("B" * 2, header_bytes))
        assert header == win_64_magic, "Unexpected header bytes in 64-bit windows object file."

    if create_ios:
        # 32-bit arm iOS mach-o files start with the following magic bytes:
        #  uint32_t []
        arm_32_ios_magic = [
            0xfeedface,  # Mach-o magic bytes
            # 0xfe, 0xed, 0xfa, 0xce, # Mach-o magic bytes
            12,  # CPU type is ARM
            11,  # CPU subtype is ARMv7s
            1]  # It's a relocatable object file.
        f = open("lesson_11_arm_32_ios.o", "rb")
        try:
            header_bytes = f.read(4 * 4)
        except:
            assert False, "ios object file not generated"
        f.close()

        header = list(unpack("I" * 4, header_bytes))
        assert header == arm_32_ios_magic, "Unexpected header bytes in 32-bit arm ios object file."

    # It looks like the object files we produced are plausible for
    # those targets. We'll count that as a success for the purposes
    # of this tutorial. For a real application you'd then need to
    # figure out how to integrate Halide into your cross-compilation
    # toolchain. There are several small examples of this in the
    # Halide repository under the apps folder. See HelloAndroid and
    # HelloiOS here:
    # https:#github.com/halide/Halide/tree/master/apps/
    print("Success!")
    return 0


if __name__ == "__main__":
    main()



import pybind11
from setuptools import find_packages
from skbuild import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README_python.md").read_text()

setup(
    name="halide",
    version='16.0.0',
    author="The Halide team",
    author_email="halide-dev@lists.csail.mit.edu",
    description="Halide is a programming language designed to make it easier "
                "to write high-performance image and array processing code.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.8",
    packages=find_packages(where="python_bindings/src"),
    package_dir={"": "python_bindings/src"},
    cmake_source_dir="python_bindings",
    cmake_args=[
        f"-Dpybind11_ROOT={pybind11.get_cmake_dir()}",
        "-DCMAKE_REQUIRE_FIND_PACKAGE_pybind11=YES",
        "-DHalide_INSTALL_PYTHONDIR=python_bindings/src",
        "-DCMAKE_INSTALL_RPATH=$<IF:$<PLATFORM_ID:Darwin>,@loader_path,$ORIGIN>",
        "-DHalide_Python_INSTALL_IMPORTED_DEPS=ON",
        "-DWITH_TESTS=NO",
        "-DWITH_TUTORIALS=NO",
        "-DWITH_PYTHON_STUBS=NO",
        "-DCMAKE_PREFIX_PATH=$ENV{CMAKE_PREFIX_PATH}",
        "--no-warn-unused-cli",
    ],
)



import halide as hl

x = hl.Var("x")
y = hl.Var("y")


@hl.generator(name="bitpy")
class BitGenerator:
    # We can use `bool` as an alias for `hl.Bool()` if we like
    bit_input = hl.InputBuffer(bool, 1)
    bit_constant = hl.InputScalar(bool)

    bit_output = hl.OutputBuffer(hl.Bool(), 1)

    def generate(self):
        g = self
        g.bit_output[x] = g.bit_input[x] | g.bit_constant


if __name__ == "__main__":
    hl.main()



import halide as hl

f32 = hl.Param(hl.Float(32), "f32", -32.0)
f64 = hl.Param(hl.Float(64), "f64", 64.0)
i16 = hl.Param(hl.Int(16), "i16", -16)
i32 = hl.Param(hl.Int(32), "i32", 32)
u16 = hl.Param(hl.UInt(16), "u16", 16)
u32 = hl.Param(hl.UInt(32), "u32", 32)


def test_types():
    # Verify that the types match the rules in match_types()
    assert (f32 / f64).type() == hl.Float(64)
    assert (f32 // f64).type() == hl.Float(64)

    assert (i16 / i32).type() == hl.Int(32)
    assert (i16 // i32).type() == hl.Int(32)

    assert (u16 / u32).type() == hl.UInt(32)
    assert (u16 // u32).type() == hl.UInt(32)

    # int / uint -> int
    assert (u16 / i32).type() == hl.Int(32)
    assert (i32 // u16).type() == hl.Int(32)

    # any / float -> float
    # float / any -> float
    assert (u16 / f32).type() == hl.Float(32)
    assert (u16 // f32).type() == hl.Float(32)

    assert (i16 / f64).type() == hl.Float(64)
    assert (i16 // f64).type() == hl.Float(64)


def test_division():
    # Verify that division semantics match those for Halide
    # (rather than python); this differs for int/int which
    # defaults to float (rather than floordiv) in Python3.
    # Also test that // always floors the result, even for float.
    assert hl.evaluate(f32 / f64) == -0.5
    assert hl.evaluate(f32 // f64) == -1.0

    assert hl.evaluate(i16 / i32) == -1
    assert hl.evaluate(i16 // i32) == -1
    assert hl.evaluate(i32 / i16) == -2

    assert hl.evaluate(u16 / u32) == 0
    assert hl.evaluate(u16 // u32) == 0

    assert hl.evaluate(u16 / i32) == 0
    assert hl.evaluate(i32 // u16) == 2

    assert hl.evaluate(u16 / f32) == -0.5
    assert hl.evaluate(u16 // f32) == -1.0

    assert hl.evaluate(i16 / f64) == -0.25
    assert hl.evaluate(i16 // f64) == -1.0


def test_division_tupled():
    # Same as test_division, but using the tuple variant
    assert hl.evaluate((f32 / f64, f32 // f64)) == (-0.5, -1.0)
    assert hl.evaluate((i16 / i32, i16 // i32, i32 / i16)) == (-1, -1, -2)
    assert hl.evaluate((u16 / u32, u16 // u32)) == (0, 0)
    assert hl.evaluate((u16 / i32, i32 // u16)) == (0, 2)
    assert hl.evaluate((u16 / f32, u16 // f32)) == (-0.5, -1.0)
    assert hl.evaluate((i16 / f64, i16 // f64)) == (-0.25, -1.0)


def test_division_gpu():
    # Allow GPU usage -- don't use f64 since not all GPU backends support that
    f = hl.cast(hl.Float(32), f64)
    assert hl.evaluate_may_gpu((f32 / f, f32 // f)) == (-0.5, -1.0)
    assert hl.evaluate_may_gpu((i16 / i32, i16 // i32, i32 / i16)) == (-1, -1, -2)
    assert hl.evaluate_may_gpu((u16 / u32, u16 // u32)) == (0, 0)
    assert hl.evaluate_may_gpu((u16 / i32, i32 // u16)) == (0, 2)
    assert hl.evaluate_may_gpu((u16 / f32, u16 // f32)) == (-0.5, -1.0)
    assert hl.evaluate_may_gpu((i16 / f, i16 // f)) == (-0.25, -1.0)


if __name__ == "__main__":
    test_types()
    test_division()
    test_division_tupled()
    test_division_gpu()



#!/usr/bin/python3

# Halide tutorial lesson 13: Tuples

# This lesson describes how to write Funcs that evaluate to multiple
# values.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_13_tuples
# in a shell with the current directory at python_bindings/

import halide as hl

import numpy as np
import math


def main():

    # So far Funcs (such as the one below) have evaluated to a single
    # scalar value for each point in their domain.
    single_valued = hl.Func()
    x, y = hl.Var("x"), hl.Var("y")
    single_valued[x, y] = x + y

    # One way to write a hl.Func that returns a collection of values is
    # to add an additional dimension which indexes that
    # collection. This is how we typically deal with color. For
    # example, the hl.Func below represents a collection of three values
    # for every x, y coordinate indexed by c.
    color_image = hl.Func()
    c = hl.Var("c")
    color_image[x, y, c] = hl.select(c == 0, 245,  # Red value
                                     c == 1, 42,  # Green value
                                     132)        # Blue value

    # Since this pattern appears quite often, Halide provides a
    # syntatic sugar to write the code above as the following,
    # using the "mux" function.
    # color_image[x, y, c] = hl.mux(c, [245, 42, 132]);

    # This method is often convenient because it makes it easy to
    # operate on this hl.Func in a way that treats each item in the
    # collection equally:
    brighter = hl.Func()
    brighter[x, y, c] = color_image[x, y, c] + 10

    # However this method is also inconvenient for three reasons.
    #
    # 1) Funcs are defined over an infinite domain, so users of this
    # hl.Func can for example access color_image(x, y, -17), which is
    # not a meaningful value and is probably indicative of a bug.
    #
    # 2) It requires a hl.select, which can impact performance if not
    # bounded and unrolled:
    # brighter.bound(c, 0, 3).unroll(c)
    #
    # 3) With this method, all values in the collection must have the
    # same type. While the above two issues are merely inconvenient,
    # this one is a hard limitation that makes it impossible to
    # express certain things in this way.

    # It is also possible to represent a collection of values as a
    # collection of Funcs:
    func_array = [hl.Func() for i in range(3)]
    func_array[0][x, y] = x + y
    func_array[1][x, y] = hl.sin(x)
    func_array[2][x, y] = hl.cos(y)

    # This method avoids the three problems above, but introduces a
    # new annoyance. Because these are separate Funcs, it is
    # difficult to schedule them so that they are all computed
    # together inside a single loop over x, y.

    # A third alternative is to define a hl.Func as evaluating to a
    # Tuple instead of an hl.Expr. A Tuple is a fixed-size collection of
    # Exprs which may have different type. The following function
    # evaluates to an integer value (x+y), and a floating point value
    # (hl.sin(x*y)).
    multi_valued = hl.Func("multi_valued")
    multi_valued[x, y] = (x + y, hl.sin(x * y))

    # Realizing a tuple-valued hl.Func returns a collection of
    # Buffers. We call this a Realization. It's equivalent to a
    # std::vector of hl.Buffer/Image objects:
    if True:
        im1, im2 = multi_valued.realize([80, 60])
        assert im1.type() == hl.Int(32)
        assert im2.type() == hl.Float(32)
        assert im1[30, 40] == 30 + 40
        assert np.isclose(im2[30, 40], math.sin(30 * 40))

    # You can also pass a tuple of pre-allocated buffers to realize()
    # rather than having new ones created. (The Buffers must have the correct
    # types and have identical sizes.)
    if True:
        im1, im2 = hl.Buffer(hl.Int(32), [80, 60]), hl.Buffer(hl.Float(32), [80, 60])
        multi_valued.realize((im1, im2))
        assert im1[30, 40] == 30 + 40
        assert np.isclose(im2[30, 40], math.sin(30 * 40))

    # All Tuple elements are evaluated together over the same domain
    # in the same loop nest, but stored in distinct allocations. The
    # equivalent C++ code to the above is:
    if True:
        multi_valued_0 = np.empty((80 * 60), dtype=np.int32)
        multi_valued_1 = np.empty((80 * 60), dtype=np.int32)

        for yy in range(80):
            for xx in range(60):
                multi_valued_0[xx + 60 * yy] = xx + yy
                multi_valued_1[xx + 60 * yy] = math.sin(xx * yy)

    # When compiling ahead-of-time, a Tuple-valued hl.Func evaluates
    # into multiple distinct output halide_buffer_t structs. These appear in
    # order at the end of the function signature:
    # int multi_valued(...input buffers and params..., halide_buffer_t
    # *output_1, halide_buffer_t *output_2)

    # You can construct a Tuple by passing multiple Exprs to the
    # Tuple constructor as we did above. Perhaps more elegantly, you
    # can also take advantage of initializer lists and just
    # enclose your Exprs in braces:
    multi_valued_2 = hl.Func("multi_valued_2")
    multi_valued_2[x, y] = (x + y, hl.sin(x * y))

    # Calls to a multi-valued hl.Func cannot be treated as Exprs. The
    # following is a syntax error:
    # hl.Func consumer
    # consumer[x, y] = multi_valued_2[x, y] + 10

    # Instead you must index the returned object with square brackets
    # to retrieve the individual Exprs:
    integer_part = multi_valued_2[x, y][0]
    floating_part = multi_valued_2[x, y][1]
    assert type(integer_part) is hl.FuncTupleElementRef
    assert type(floating_part) is hl.FuncTupleElementRef

    consumer = hl.Func()
    consumer[x, y] = (integer_part + 10, floating_part + 10.0)

    # Tuple reductions.
    if True:
        # Tuples are particularly useful in reductions, as they allow
        # the reduction to maintain complex state as it walks along
        # its domain. The simplest example is an argmax.

        # First we create an Image to take the argmax over.
        input_func = hl.Func()
        input_func[x] = hl.sin(x)
        input = input_func.realize([100])
        assert input.type() == hl.Float(32)

        # Then we defined a 2-valued Tuple which tracks the maximum value
        # its index.
        arg_max = hl.Func()

        # Pure definition.
        # (using [()] for zero-dimensional Funcs is a convention of this python interface)
        arg_max[()] = (0, input[0])

        # Update definition.
        r = hl.RDom([(1, 99)])
        old_index = arg_max[()][0]
        old_max = arg_max[()][1]
        new_index = hl.select(old_max > input[r], r, old_index)
        new_max = hl.max(input[r], old_max)
        arg_max[()] = (new_index, new_max)

        # The equivalent C++ is:
        arg_max_0 = 0
        arg_max_1 = float(input[0])
        for r in range(1, 100):
            old_index = arg_max_0
            old_max = arg_max_1
            new_index = r if (old_max > input[r]) else old_index
            new_max = max(input[r], old_max)
            # In a tuple update definition, all loads and computation
            # are done before any stores, so that all Tuple elements
            # are updated atomically with respect to recursive calls
            # to the same hl.Func.
            arg_max_0 = new_index
            arg_max_1 = new_max

        # Let's verify that the Halide and C++ found the same maximum
        # value and index.
        if True:
            r0, r1 = arg_max.realize()

            assert r0.type() == hl.Int(32)
            assert r1.type() == hl.Float(32)
            assert arg_max_0 == r0[()]
            assert np.isclose(arg_max_1, r1[()])

        # Halide provides argmax and hl.argmin as built-in reductions
        # similar to sum, product, maximum, and minimum. They return
        # a Tuple consisting of the point in the reduction domain
        # corresponding to that value, and the value itself. In the
        # case of ties they return the first value found. We'll use
        # one of these in the following section.

    # Tuples for user-defined types.
    if True:
        # Tuples can also be a convenient way to represent compound
        # objects such as complex numbers. Defining an object that
        # can be converted to and from a Tuple is one way to extend
        # Halide's type system with user-defined types.
        class Complex:

            def __init__(self, r, i=None):
                if type(r) is float and type(i) is float:
                    self.real = hl.Expr(r)
                    self.imag = hl.Expr(i)
                elif i is not None:
                    self.real = r
                    self.imag = i
                else:
                    self.real = r[0]
                    self.imag = r[1]

            def as_tuple(self):
                "Convert to a Tuple"
                return (self.real, self.imag)

            def __add__(self, other):
                "Complex addition"
                return Complex(self.real + other.real, self.imag + other.imag)

            def __mul__(self, other):
                "Complex multiplication"
                return Complex(self.real * other.real - self.imag * other.imag,
                               self.real * other.imag + self.imag * other.real)

            def __getitem__(self, idx):
                return (self.real, self.imag)[idx]

            def __len__(self):
                return 2

            def magnitude(self):
                "Complex magnitude"
                return (self.real * self.real) + (self.imag * self.imag)

            # Other complex operators would go here. The above are
            # sufficient for this example.

        # Let's use the Complex struct to compute a Mandelbrot set.
        mandelbrot = hl.Func()

        # The initial complex value corresponding to an x, y coordinate
        # in our hl.Func.
        initial = Complex(x / 15.0 - 2.5, y / 6.0 - 2.0)

        # Pure definition.
        t = hl.Var("t")
        mandelbrot[x, y, t] = Complex(0.0, 0.0)

        # We'll use an update definition to take 12 steps.
        r = hl.RDom([(1, 12)])
        current = Complex(mandelbrot[x, y, r - 1])

        # The following line uses the complex multiplication and
        # addition we defined above.
        mandelbrot[x, y, r] = (Complex(current * current) + initial)

        # We'll use another tuple reduction to compute the iteration
        # number where the value first escapes a circle of radius 4.
        # This can be expressed as an hl.argmin of a boolean - we want
        # the index of the first time the given boolean expression is
        # false (we consider false to be less than true).  The argmax
        # would return the index of the first time the expression is
        # true.

        escape_condition = Complex(mandelbrot[x, y, r]).magnitude() < 16.0
        first_escape = hl.argmin(escape_condition)
        assert type(first_escape) is tuple
        # We only want the index, not the value, but hl.argmin returns
        # both, so we'll index the hl.argmin Tuple expression using
        # square brackets to get the hl.Expr representing the index.
        escape = hl.Func()
        escape[x, y] = first_escape[0]

        # Realize the pipeline and print the result as ascii art.
        result = escape.realize([61, 25])
        assert result.type() == hl.Int(32)
        code = " .:-~*={&%#@"
        for yy in range(result.height()):
            for xx in range(result.width()):
                index = result[xx, yy]
                if index < len(code):
                    print("%c" % code[index], end="")
                else:
                    pass  # is lesson 13 cpp version buggy ?
            print("")

    print("Success!")

    return 0


if __name__ == "__main__":
    main()



"""
Simple blur.
"""

import halide as hl
import enum


class BlurGPUSchedule(enum.Enum):
    # Fully inlining schedule.
    Inline = 0
    # Schedule caching intermedia result of blur_x.
    Cache = 1
    # Schedule enabling sliding window opt within each work-item or cuda
    # thread.
    Slide = 2
    # The same as above plus vectorization per work-item.
    SlideVectorize = 3


_GPU_SCHEDULE_ENUM_MAP = {
    "inline": BlurGPUSchedule.Inline,
    "cache": BlurGPUSchedule.Cache,
    "slide": BlurGPUSchedule.Slide,
    "slide_vector": BlurGPUSchedule.SlideVectorize,
}


@hl.generator()
class blur:
    gpu_schedule = hl.GeneratorParam("slide_vector")
    gpu_tile_x = hl.GeneratorParam(32)
    gpu_tile_y = hl.GeneratorParam(8)

    # Note: although this is declared as operating on uint16 images,
    # it will produce incorrect results if more than 14-bit images are used.
    input_buf = hl.InputBuffer(hl.UInt(16), 2)
    blur_y = hl.OutputBuffer(hl.UInt(16), 2)

    def generate(self):
        g = self

        x, y, xi, yi = hl.vars("x y xi yi")

        # The algorithm
        clamped = hl.BoundaryConditions.repeat_edge(g.input_buf)

        blur_x = hl.Func("blur_x")
        blur_x[x, y] = (clamped[x, y] + clamped[x + 1, y] + clamped[x + 2, y]) // 3
        g.blur_y[x, y] = (blur_x[x, y] + blur_x[x, y + 1] + blur_x[x, y + 2]) // 3

        # How to schedule it
        if g.target().has_gpu_feature():
            # GPU schedule.

            # This will raise an exception for unknown strings, which is what
            # we want
            schedule_enum = _GPU_SCHEDULE_ENUM_MAP[g.gpu_schedule]

            if schedule_enum == BlurGPUSchedule.Inline:
                # - Fully inlining.
                g.blur_y.gpu_tile(x, y, xi, yi, g.gpu_tile_x, g.gpu_tile_y)

            elif schedule_enum == BlurGPUSchedule.Cache:
                # - Cache blur_x calculation.
                g.blur_y.gpu_tile(x, y, xi, yi, g.gpu_tile_x, g.gpu_tile_y)
                blur_x.compute_at(g.blur_y, x).gpu_threads(x, y)

            elif schedule_enum == BlurGPUSchedule.Slide:
                # - Instead of caching blur_x calculation explicitly, the
                #   alternative is to allow each work-item in OpenCL or thread
                #   in CUDA to calculate more rows of blur_y so that temporary
                #   blur_x calculation is re-used implicitly. This achieves
                #   the similar schedule of sliding window.
                y_inner = hl.Var("y_inner")
                (
                    g.blur_y.split(y, y, y_inner, g.gpu_tile_y)
                    .reorder(y_inner, x)
                    .unroll(y_inner)
                    .gpu_tile(x, y, xi, yi, g.gpu_tile_x, 1)
                )

            elif schedule_enum == BlurGPUSchedule.SlideVectorize:
                # Vectorization factor.
                factor = 2
                y_inner = hl.Var("y_inner")
                (
                    g.blur_y.vectorize(x, factor)
                    .split(y, y, y_inner, g.gpu_tile_y)
                    .reorder(y_inner, x)
                    .unroll(y_inner)
                    .gpu_tile(x, y, xi, yi, g.gpu_tile_x, 1)
                )

        elif g.target().has_feature(hl.TargetFeature.HVX):
            # Hexagon schedule.
            # TODO: Try using a schedule like the CPU one below.
            vector_size = 128

            (
                g.blur_y.compute_root()
                .hexagon()
                .prefetch(g.input_buf, y, y, 2)
                .split(y, y, yi, 128)
                .parallel(y)
                .vectorize(x, vector_size * 2)
            )
            (
                blur_x.store_at(g.blur_y, y)
                .compute_at(g.blur_y, yi)
                .vectorize(x, vector_size)
            )
        else:
            # CPU schedule.
            # Compute blur_x as needed at each vector of the output.
            # Halide will store blur_x in a circular buffer so its
            # results can be re-used.
            vector_size = g.natural_vector_size(g.input_buf.type())
            g.blur_y.split(y, y, yi, 32).parallel(y).vectorize(x, vector_size)
            (
                blur_x.store_at(g.blur_y, y)
                .compute_at(g.blur_y, x)
                .vectorize(x, vector_size)
            )


if __name__ == "__main__":
    hl.main()



import array
from user_context import user_context


def test():
    output = bytearray("\0\0\0\0", "ascii")
    user_context(None, ord("q"), output)
    assert output == bytearray("qqqq", "ascii")


if __name__ == "__main__":
    test()



import halide as hl

x = hl.Var("x")
y = hl.Var("y")
c = hl.Var("c")

# Note that this Generator explicitly inherits from hl.Generator;
# this isn't necessary (the decorator will inject the base class for you),
# but is allowed for better compatibility with static type checkers such
# as pytype. Thus, this is left here as a test to verify that this option
# remains functional.


@hl.generator(name="complexpy")
class ComplexPy(hl.Generator):
    vectorize = hl.GeneratorParam(True)
    extra_input_name = hl.GeneratorParam("")

    typed_buffer_input = hl.InputBuffer(hl.UInt(8), 3)
    untyped_buffer_input = hl.InputBuffer(None, 3)
    simple_input = hl.InputBuffer(None, 3)
    float_arg = hl.InputScalar(hl.Float(32))
    # We can use `int` as an alias for `hl.Int(32)` if we like
    int_arg = hl.InputScalar(int)

    simple_output = hl.OutputBuffer(hl.Float(32), 3)
    tuple_output = hl.OutputBuffer(None, 3)
    typed_buffer_output = hl.OutputBuffer(hl.Float(32), 3)
    untyped_buffer_output = hl.OutputBuffer(None, None)
    static_compiled_buffer_output = hl.OutputBuffer(hl.UInt(8), 3)
    scalar_output = hl.OutputScalar(hl.Float(32))

    def configure(self):
        g = self
        if len(g.extra_input_name):
            g.add_input(g.extra_input_name, hl.InputBuffer(hl.UInt(16), 3))
        g.add_output("extra_output", hl.OutputBuffer(hl.Float(64), 2))

    def generate(self):
        g = self

        g.simple_output[x, y, c] = hl.f32(g.simple_input[x, y, c])
        g.typed_buffer_output[x, y, c] = hl.f32(g.typed_buffer_input[x, y, c])
        g.untyped_buffer_output[x, y, c] = hl.cast(
            g.untyped_buffer_output.type(),
            g.untyped_buffer_input[x, y, c],
        )

        intermediate = hl.Func("intermediate")
        intermediate[x, y, c] = g.simple_input[x, y, c] * g.float_arg

        g.tuple_output[x, y, c] = (
            intermediate[x, y, c],
            intermediate[x, y, c] + g.int_arg,
        )

        # This should be compiled into the Generator product itself,
        # and not produce another input for the Stub or AOT filter.
        static_compiled_buffer = hl.Buffer(hl.UInt(8), [4, 4, 1])
        for xx in range(4):
            for yy in range(4):
                for cc in range(1):
                    static_compiled_buffer[xx, yy, cc] = xx + yy + cc + 42

        g.static_compiled_buffer_output[x, y, c] = static_compiled_buffer[x, y, c]

        extra_input = getattr(g, g.extra_input_name, None)
        if extra_input:
            g.extra_output[x, y] = hl.f64(extra_input[x, y, 0] + 1)
        else:
            g.extra_output[x, y] = hl.f64(0)

        g.scalar_output[()] = g.float_arg + g.int_arg

        (
            intermediate.compute_at(g.tuple_output, y)
            .specialize(g.vectorize)
            .vectorize(x, g.natural_vector_size(hl.Float(32)))
        )


if __name__ == "__main__":
    hl.main()



# Halide tutorial lesson 2.

# This lesson demonstrates how to pass in input images.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_02_input_image
# in a shell with the current directory at python_bindings/

import halide as hl
import numpy as np
import halide.imageio
import os.path


def main():

    # This program defines a single-stage imaging pipeline that
    # brightens an image.

    # First we'll load the input image we wish to brighten.
    image_path = os.path.join(os.path.dirname(__file__), "../../tutorial/images/rgb.png")

    # We create a hl.Buffer object to wrap the numpy array
    input = hl.Buffer(halide.imageio.imread(image_path))
    assert input.type() == hl.UInt(8)

    # Next we define our hl.Func object that represents our one pipeline
    # stage.
    brighter = hl.Func("brighter")

    # Our hl.Func will have three arguments, representing the position
    # in the image and the color channel. Halide treats color
    # channels as an extra dimension of the image.
    x, y, c = hl.Var("x"), hl.Var("y"), hl.Var("c")

    # Normally we'd probably write the whole function definition on
    # one line. Here we'll break it apart so we can explain what
    # we're doing at every step.

    # For each pixel of the input image.
    value = input[x, y, c]
    assert type(value) == hl.Expr

    # Cast it to a floating point value.
    value = hl.cast(hl.Float(32), value)

    # Multiply it by 1.5 to brighten it. Halide represents real
    # numbers as floats, not doubles, so we stick an 'f' on the end
    # of our constant.
    value = value * 1.5

    # Clamp it to be less than 255, so we don't get overflow when we
    # hl.cast it back to an 8-bit unsigned int.
    value = hl.min(value, 255.0)

    # Cast it back to an 8-bit unsigned integer.
    value = hl.cast(hl.UInt(8), value)

    # Define the function.
    brighter[x, y, c] = value

    # The equivalent one-liner to all of the above is:
    #
    # brighter(x, y, c) = hl.cast<uint8_t>(hl.min(input(x, y, c) * 1.5f, 255))
    # brighter[x, y, c] = hl.cast(hl.UInt(8), hl.min(input[x, y, c] * 1.5, 255))
    #
    # In the shorter version:
    # - I skipped the hl.cast to float, because multiplying by 1.5f does
    #   that automatically.
    # - I also used integer constants in hl.clamp, because they get hl.cast
    #   to match the type of the first argument.
    # - I left the h. off hl.clamp. It's unnecessary due to Koenig
    #   lookup.

    # Remember. All we've done so far is build a representation of a
    # Halide program in memory. We haven't actually processed any
    # pixels yet. We haven't even compiled that Halide program yet.

    # So now we'll realize the hl.Func. The size of the output image
    # should match the size of the input image. If we just wanted to
    # brighten a portion of the input image we could request a
    # smaller size. If we request a larger size Halide will throw an
    # error at runtime telling us we're trying to read out of bounds
    # on the input image.
    output_image = brighter.realize([input.width(), input.height(), input.channels()])
    assert output_image.type() == hl.UInt(8)

    # Save the output for inspection. It should look like a bright parrot.
    # python3-imageio versions <2.5 expect a numpy array
    halide.imageio.imwrite("brighter.png", np.asanyarray(output_image))
    print("Created brighter.png result file.")

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Halide tutorial lesson 1.

# This lesson demonstrates basic usage of Halide as a JIT compiler for imaging.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_01_basics
# in a shell with the current directory at python_bindings/

import halide as hl


def main():

    # This program defines a single-stage imaging pipeline that
    # outputs a grayscale diagonal gradient.

    # A 'hl.Func' object represents a pipeline stage. It's a pure
    # function that defines what value each pixel should have. You
    # can think of it as a computed image.
    gradient = hl.Func("gradient")

    # hl.Var objects are names to use as variables in the definition of
    # a hl.Func. They have no meaning by themselves.
    x, y = hl.Var("x"), hl.Var("y")

    # We typically use Vars named 'x' and 'y' to correspond to the x
    # and y axes of an image, and we write them in that order. If
    # you're used to thinking of images as having rows and columns,
    # then x is the column index, and y is the row index.

    # Funcs are defined at any integer coordinate of its variables as
    # an hl.Expr in terms of those variables and other functions.
    # Here, we'll define an hl.Expr which has the value x + y. Vars have
    # appropriate operator overloading so that expressions like
    # 'x + y' become 'hl.Expr' objects.
    e = x + y
    assert type(e) == hl.Expr

    # Now we'll add a definition for the hl.Func object. At pixel x, y,
    # the image will have the value of the hl.Expr e. On the left hand
    # side we have the hl.Func we're defining and some Vars. On the right
    # hand side we have some hl.Expr object that uses those same Vars.
    gradient[x, y] = e

    # This is the same as writing:
    #
    #   gradient[x, y] = x + y
    #
    # which is the more common form, but we are showing the
    # intermediate hl.Expr here for completeness.

    # That line of code defined the hl.Func, but it didn't actually
    # compute the output image yet. At this stage it's just Funcs,
    # Exprs, and Vars in memory, representing the structure of our
    # imaging pipeline. We're meta-programming. This C++ program is
    # constructing a Halide program in memory. Actually computing
    # pixel data comes next.

    # Now we 'realize' the hl.Func, which JIT compiles some code that
    # implements the pipeline we've defined, and then runs it.  We
    # also need to tell Halide the domain over which to evaluate the
    # hl.Func, which determines the range of x and y above, and the
    # resolution of the output image. Halide.h also provides a basic
    # templatized image type we can use. We'll make an 800 x 600
    # image.
    output = gradient.realize([800, 600])
    assert output.type() == hl.Int(32)

    # Halide does type inference for you. hl.Var objects represent
    # 32-bit integers, so the hl.Expr object 'x + y' also represents a
    # 32-bit integer, and so 'gradient' defines a 32-bit image, and
    # so we got a 32-bit signed integer image out when we call
    # 'realize'. Halide types and type-casting rules are equivalent
    # to C.

    # Let's check everything worked, and we got the output we were
    # expecting:
    for j in range(output.height()):
        for i in range(output.width()):
            # We can access a pixel of an hl.Buffer object using similar
            # syntax to defining and using functions.
            assert output[i, j] == i + j, \
                "Something went wrong!\n" + \
                "Pixel %d, %d was supposed to be %d, but instead it's %d\n" % (
                    i, j, i + j, output[i, j])

    # Everything worked! We defined a hl.Func, then called 'realize' on
    # it to generate and run machine code that produced a hl.Buffer.
    print("Success!")

    return 0

if __name__ == "__main__":
    main()
